"""Module for consuming the streams of a task through a websocket.

Task logs are output in the following format:

$ induction logs <task_id>
<log line 1>                                        log content
<log line 2>
  ...
<log line N>
                                                    `SPACING` black lines
 ● 0:00:02 Streaming output from task <task_id>     connection status footer
"""
from threading import Thread, RLock
from datetime import timedelta
from typing import IO
import logging
import sched
import json
import sys
import time
import os

import websocket

import inductiva
from inductiva import constants

logger = logging.getLogger("websocket")
logger.setLevel(logging.ERROR)

SPACING = 1
ESC = "\u001B["
RESET = f"{ESC}0m"
EMPHASIS = f"{ESC}2m"
FAILURE = f"{ESC}31m"
SUCCESS = f"{ESC}32m"
PENDING = f"{ESC}33m"
CLEAR_LINE = f"{ESC}0K"
UP = f"{ESC}{SPACING}A"
DOWN = f"{ESC}{SPACING}B"

ANSI_ENABLED = os.getenv("ANSI_ENABLED", "1") == "1"


class TaskStreamConsumer:
    """Consumer of the streams of a running task through a websocket.

    Attributes:
        RECONNECT_DELAY_SEC (float): Delay interval when reconnecting in
            seconds.
        PING_INTERVAL_SEC (float): Time interval, in seconds, to automatically
            send a “ping” command  to the websocket (in seconds).
            If set to 0, no ping is sent periodically.
        PING_TIMEOUT_SEC (float): Timeout (in seconds) to close the stream
            if the pong message is not received.
    """
    RECONNECT_DELAY_SEC = 5.
    PING_INTERVAL_SEC = 15.
    PING_TIMEOUT_SEC = 5.
    TICK_INTERVAL_SEC = 1.

    def __init__(self,
                 task_id: str,
                 fout: IO = sys.stdout,
                 ferr: IO = sys.stderr):
        """Initialize websocket connection to the task STDOUT & STDERR streams.

        Args:
            task_id (int): ID of the task for which to get the STDOUT & STDERR
                streams.
            fout (IO): I/O streams, such as returned by open(), for the STDOUT.
            ferr (IO): I/O streams, such as returned by open(), for the STDERR.
        """
        query = f'?query={{task_id="{task_id}"}}'
        websocket_url = f"{constants.LOGS_WEBSOCKET_URL}/loki/api/v1/tail"
        self.websocket_url = websocket_url + query
        self.task_id = task_id
        self.fout = fout
        self.ferr = ferr

        # Setup the websocket application to connect to the logs socket.
        self.ws = self.__setup_websocket()

        # Check if the file handles are interactive and if ANSI is enabled.
        # Note that there is no point in using ANSI if ferr is not interactive
        self._ferr_isatty = ANSI_ENABLED and self.ferr.isatty()
        self._fout_isatty = ANSI_ENABLED and self.fout.isatty()
        self._fout_isatty &= self._ferr_isatty
        self._message_formatter = self._get_message_formatter()
        self._status_formatter = self._get_status_formatter()

        # housekeeping variables
        self._conn_start_time = None
        self._conn_alive_sec = 0.0
        self._conn_opened = False
        self._prev_status = None
        self._conn_count = 0
        self._lock = RLock()

    def __setup_websocket(self):
        """Initialize the websocket app, with the callbacks within the class."""
        return websocket.WebSocketApp(self.websocket_url,
                                      header={"X-API-Key": inductiva.api_key},
                                      on_open=self.__on_open,
                                      on_error=self.__on_error,
                                      on_close=self.__on_close,
                                      on_message=self.__on_message)

    def __on_message(self, unused_ws, message):
        data = json.loads(message)
        streams = data.get("streams")
        for stream in streams:
            msg = stream["values"][0][1]
            self._write_message(msg)

    def _get_message_formatter(self):
        """Get the message formatter based on the output file
        handle interactivity.
        """
        if self._fout_isatty:
            return self._ansi_message_formatter
        return self._plain_message_formatter

    def _get_status_formatter(self):
        """Get the message formatter based on the error file
        handle interactivity.
        """
        if self._ferr_isatty:
            return self._ansi_status_formatter
        return self._plain_status_formatter

    def _ansi_message_formatter(self, msg):
        """Fancy message formatter for formatted message output using
        ANSI escape codes.
        """
        # 1. move up SPACING lines
        # 2. move to beginning of line and clear line
        # 3. write message
        # 4. add new line
        # 5. move down SPACING lines
        # 6. move to beginning of line and clear line
        # 7. add new blank line
        return f"{UP}\r{CLEAR_LINE}{msg}\n{DOWN}\r{CLEAR_LINE}\n"

    def _plain_message_formatter(self, msg):
        """Simple message formatter for plain message output."""
        return msg + "\n"

    def _ansi_status_formatter(self, msg, status):
        """Fancy message formatter for formatted status message output using
        ANSI escape codes.
        """
        if self._conn_start_time is None:
            elapsed = ""
        else:
            alive = timedelta(seconds=int(self._conn_alive_sec))
            elapsed = f"{str(alive)} "

        # 1. go to beginning of line
        # 2. clear line
        # 3. set color for bullet
        #   3.1 write bullet icon
        #   3.2 reset color
        # 4. set emphasis formatter
        #   4.1. write timer message (if any)
        #   4.2. write status message
        # 5. reset everything
        s = f"\r{CLEAR_LINE}{status} ● {RESET}{EMPHASIS}{elapsed}{msg}{RESET}"
        return s

    def _plain_status_formatter(self, msg, _):
        """Simple message formatter for plain status message output."""
        return msg + "\n"

    def _write_message(self, msg):
        """Write the message to the output file handle

        Args:
            msg (str): message to be written to output.
        """
        s = self._message_formatter(msg)
        self.fout.write(s)
        self.fout.flush()
        self._redraw_status()

    def _write_status(self, msg, status):
        """Write the status message to the error file handle

        Args:
            msg (str): Status message.
            status (str): Status formatting code.
        """
        s = self._status_formatter(msg, status)
        self.ferr.write(s)
        self.ferr.flush()

    def _update_status(self, msg, status):
        """Update the connection status message.

        Args:
            msg (str): Status message.
            status (str): Status formatting code.
        """
        with self._lock:
            self._prev_status = (msg, status)
            self._write_status(msg, status)

    def _redraw_status(self):
        """Redraw the connection status message."""
        if self._prev_status and self._ferr_isatty:
            self._update_status(*self._prev_status)

    def __on_error(self, unused_ws, error):
        """Callback for websocket error event."""
        if not isinstance(error, KeyboardInterrupt):
            self._update_status(str(error), FAILURE)

    def __on_close(self, unused_ws, close_status_code, close_message):
        """Callback for websocket close event."""

        if close_status_code or close_message:
            close_msg = f" (code: {close_status_code}, msg: {close_message})"
        else:
            close_msg = ""

        if self._conn_opened:
            msg = f"Disconnected from task {self.task_id}{close_msg}."
            status = SUCCESS
        else:
            msg = f"Failed to connect to task {self.task_id}{close_msg}."
            status = FAILURE

        self._update_status(msg, status)
        self._conn_start_time = None
        self._conn_opened = False

    def __on_open(self, unused_ws):
        """Callback for websocket open event."""
        self._conn_start_time = time.time()
        self._conn_opened = True
        self._conn_count += 1

        msg = f"Streaming output from task {self.task_id}"
        if ANSI_ENABLED and not self._fout_isatty:
            msg += " (redirected)"

        self._update_status(msg, SUCCESS)

        # if the connection was re-established, warn the user about
        # the possiblity of some log lines being missing
        if self._conn_count > 1:
            self._write_message(" --- Connection re-established: "
                                "some log lines might be missing --- ")

    def __init_status(self):
        """Allocate sufficient lines for the status message."""
        if self._ferr_isatty:
            self.ferr.write(SPACING * "\n")
            self.ferr.flush()

    def run_forever(self):
        """Stream the STDOUT & STDERR of a task through the websocket."""

        msg = f"Opening connection to task {self.task_id}..."
        self.__init_status()
        self._update_status(msg, PENDING)
        time.sleep(2)

        if scheduler := self._get_ontick_scheduler():
            thread = Thread(target=scheduler.run)
            thread.daemon = True
            thread.start()

        self.ws.run_forever(reconnect=self.RECONNECT_DELAY_SEC,
                            ping_interval=self.PING_INTERVAL_SEC,
                            ping_timeout=self.PING_TIMEOUT_SEC)

    def _get_ontick_scheduler(self):
        """Get a scheduler for the connection alive ticker.

        Gets a scheduler that runs every second to update the time elapsed
        since the connection was opened.
        """
        if not self._ferr_isatty:
            return None

        scheduler = sched.scheduler(time.time, time.sleep)
        scheduler.enter(self.TICK_INTERVAL_SEC, 1, self._ontick, (scheduler,))
        return scheduler

    def _ontick(self, scheduler):
        """Update the connection alive time every second.

        Updates the time elapsed since the connection was opened, redraws
        the status message and schedules the next update.
        """
        if self._conn_start_time is not None:
            self._conn_alive_sec = time.time() - self._conn_start_time
            self._redraw_status()

        # schedule the next update
        scheduler.enter(self.TICK_INTERVAL_SEC, 1, self._ontick, (scheduler,))



#pylint: disable=missing-module-docstring
from inductiva.tasks.task import Task
from inductiva.tasks.run_simulation import run_simulation
from inductiva.tasks.methods import get, to_dict
from . import streams



"""Methods to interact with the tasks submitted to the API."""
from collections import defaultdict
import json
from typing import Any, Dict, Iterable, List, Mapping, Optional, Union

import inductiva
from inductiva import api
from inductiva.client import ApiClient, ApiException
from inductiva.client.apis.tags.tasks_api import TasksApi
from inductiva.client import models
from inductiva.tasks.task import Task


def to_dict(list_of_tasks: Iterable[Task]) -> Mapping[str, List[Any]]:
    """
    Converts an Iterable of tasks to a dictionary with all the
    relevant information for all the tasks.
        Args:
            list_of_tasks: An Iterable of tasks.
        Returns:
            A dictionary with all the relevant information for
            all the tasks. Example: { "ID": [1, 2, 3], 
            "Simulator": ["reef3d", "reef3d", "reef3d"], ... }
    """
    column_names = [
        "ID", "Simulator", "Status", "Submitted", "Started", "Computation Time",
        "Resource Type"
    ]
    table = defaultdict(list, {key: [] for key in column_names})

    for task in list_of_tasks:
        info = task.get_info()
        status = task.get_status()
        computation_end_time = info.get("computation_end_time", None)
        execution_time = task.get_computation_time(fail_if_running=False)

        if execution_time is not None:
            if computation_end_time is None:
                if status in ["started", "submitted"]:
                    execution_time = f"*{execution_time}"
                else:
                    execution_time = "n/a"

        executer = info["executer"]
        if executer is None:
            resource_type = None
        else:
            resource_type = executer["vm_type"]
            n_mpi_hosts = executer["n_mpi_hosts"]
            if n_mpi_hosts > 1:
                resource_type += f" x{n_mpi_hosts}"
        table["ID"].append(task.id)
        table["Simulator"].append(task.get_simulator_name())
        table["Status"].append(status)
        table["Submitted"].append(info.get("input_submit_time", None))
        table["Started"].append(info.get("start_time", None))
        table["Computation Time"].append(execution_time)
        table["Resource Type"].append(resource_type)
    return table


def _fetch_tasks_from_api(
    status: Optional[Union[str, models.TaskStatusCode]] = None,
    page=1,
    per_page=10,
) -> List[Dict]:
    """Get information about a user's tasks on the API.

    Tags can be filtered by a status. Results are paginated indexed from 1.
    """
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = TasksApi(client)

        query_params = {
            "page": page,
            "per_page": per_page,
        }

        if status is not None:
            query_params["status"] = models.TaskStatusCode(status)

        try:
            # Get User Tasks
            resp = api_instance.get_user_tasks(
                query_params=query_params,
                skip_deserialization=True,  # avoid deserializing to model,
                # leave as dict which we'll later serialize to our own
                # dataclasses
            ).response

            response_body = json.loads(resp.data.decode("utf-8"))

            return [{**task} for task in response_body]

        except ApiException as e:
            raise e


def get(
    last_n: int = 5,
    status: Optional[Union[str, models.TaskStatusCode]] = None
) -> List["inductiva.tasks.Task"]:
    """Get the last N tasks of a user.

    This function fetches info about the last N tasks (with respect to
    submission time) of a user to stdout, sorted by submission time with the
    most recent first.
    A status can be specified to filter to get only tasks with that status, in
    which case the last N tasks with that status will be listed.
    The number of tasks can be less than N if the aren't enough tasks that match
    the specified criteria.

    Similar to the inductiva.task.list() function, but instead of printing
    to stdout, returns a list of Task objects which can be used to perform
    further operations on those tasks (e.g., download outputs,
    kill unfinished tasks, etc.).

    Example usage:
        # get the last 5 tasks that haven't started yet and kill them
        tasks = inductiva.tasks.get(5, status="submitted")
        for task in tasks:
            task.kill()

    Args:
        last_n: The number of most recent tasks with respect to submission
            time to fetch. If filtering criteria (currently status is available)
            is specified, most recent N tasks that match that criteria will be
            listed. The actual number of tasks may be less if there
            aren't enough tasks available.
        status: The status of the tasks to get. If None, tasks with any status
            will be returned.

    Returns:
        List of Task objects.
    """
    if last_n < 1:
        raise ValueError("last_n must be >= 1")

    status = models.TaskStatusCode(status) if status is not None else None

    raw_tasks_info = _fetch_tasks_from_api(status, page=1, per_page=last_n)
    tasks = [
        inductiva.tasks.Task.from_api_info(info) for info in raw_tasks_info
    ]

    return tasks


def get_all(
        status: Optional[Union[str,
                               models.TaskStatusCode]] = None) -> List[Dict]:
    """Get all tasks of a user.

    This function fetches all tasks of a user, sorted by submission
    time with the most recent first. If status is specified, only
    tasks with that status will be fetched.
    Args:
        status: The status of the tasks to get. If None, tasks with any status
            will be returned.
    Returns:
        List of dictionaries with information about the tasks.
    """
    status = models.TaskStatusCode(status) if status is not None else None

    all_tasks = []
    page_counter = 1

    while tasks_fetched := _fetch_tasks_from_api(status,
                                                 page=page_counter,
                                                 per_page=50):
        all_tasks.extend(tasks_fetched)
        page_counter += 1
    return all_tasks



"""Main CLI entrypoint."""
import argparse
import os

import inductiva
from inductiva import _cli
from inductiva import constants
from . import loader


def main():
    parser = argparse.ArgumentParser(
        prog="inductiva",
        description="CLI tool for Inductiva API.",
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {inductiva.__version__}",
    )

    parser.add_argument(
        "--api-key",
        type=str,
        help=("API key to use. If not provided, it "
              "will be read from the INDUCTIVA_API_KEY environment variable."),
    )

    # If no subcommand is provided, print help
    _cli.utils.show_help_msg(parser)

    # Create subcommands
    subparsers = parser.add_subparsers(title="available subcomands")

    # Load all modules starting with "cmd_" as subcommands.
    loader.load_commands(subparsers,
                         os.path.dirname(__file__),
                         "inductiva._cli",
                         prefix=constants.LOADER_COMMAND_PREFIX)

    args = parser.parse_args()
    if args.api_key:
        inductiva.api_key = args.api_key

    # Call the function associated with the subcommand
    try:
        exit_code = args.func(args)
    except Exception as e:  # pylint: disable=broad-except
        exit_code = 1
        print(e)
    return exit_code



"""Manage running/completed tasks on the Inductiva API."""
import pathlib
import contextlib
import time
import json
from absl import logging
from typing import Dict, Any, List, Optional, Tuple, Union
from typing_extensions import TypedDict
import datetime
from ..localization import translator as __

from inductiva import constants
from inductiva.client import exceptions, models
from inductiva import api, types
from inductiva.client.apis.tags import tasks_api
from inductiva.utils import files, format_utils, data, output_contents

import warnings


class Task:
    """Represents a running/completed task on the Inductiva API.

    Example usage:
        task = scenario.simulate(...)
        final_status = task.wait()
        info = task.get_info() # dictionary with info about the task
        task.download_outputs(
            filenames=["file1.txt", "file2.dat"] # download only these files
        )

    Attributes:
        id: The task ID.
        _api: Instance of TasksApi (from the generated API client).
    """

    FAILED_STATUSES = {
        models.TaskStatusCode.FAILED, models.TaskStatusCode.KILLED,
        models.TaskStatusCode.EXECUTERFAILED,
        models.TaskStatusCode.EXECUTERTERMINATED,
        models.TaskStatusCode.EXECUTERTERMINATEDBYUSER,
        models.TaskStatusCode.SPOTINSTANCEPREEMPTED,
        models.TaskStatusCode.ZOMBIE
    }

    TERMINAL_STATUSES = {models.TaskStatusCode.SUCCESS}.union(FAILED_STATUSES)

    RUNNING_STATUSES = {
        models.TaskStatusCode.PENDINGINPUT, models.TaskStatusCode.STARTED
    }

    KILLABLE_STATUSES = {models.TaskStatusCode.SUBMITTED
                        }.union(RUNNING_STATUSES)

    KILL_VERBOSITY_LEVELS = [0, 1, 2]

    def __init__(self, task_id: str):
        """Initialize the instance from a task ID."""
        self.id = task_id
        self._api = tasks_api.TasksApi(api.get_client())
        self._info = None
        self._status = None

    def is_running(self) -> bool:
        """Validate if the task is running.

        This method issues a request to the API.
        """
        return self.get_status() == models.TaskStatusCode.STARTED

    def is_failed(self) -> bool:
        """Validate if the task has failed.
        
        This method issues a request to the API.
        """
        return self.get_status() in self.FAILED_STATUSES

    def is_terminal(self) -> bool:
        """Check if the task is in a terminal status.

        This method issues a request to the API.
        """
        return self.get_status() in self.TERMINAL_STATUSES

    @classmethod
    def from_api_info(cls, info: Dict[str, Any]) -> "Task":

        task = cls(info["task_id"])
        task._info = info
        task._status = models.TaskStatusCode(info["status"])

        # TODO(luispcunha): construct correct output class from API info.

        return task

    @contextlib.contextmanager
    def sync_context(self):
        """Enter context manager for blocking sync execution.

        This turns an asynchronous task into a blocking task.
        If an exception/ctrl+c is caught while in the context manager, the
        remote task is killed.

        Usage:
            task = scenario.simulate(...)

            with task.sync_context():
                # If an exception happens here or ctrl+c is pressed, the task
                # will be killed.
                task.wait()

        """
        try:
            yield
        except KeyboardInterrupt:
            logging.info("Caught SIGINT: terminating blocking task...")
            self.kill()
        except Exception as e:
            logging.error("Caught exception: terminating blocking task...")
            self.kill()
            raise e

    def get_status(self) -> models.TaskStatusCode:
        """Get status of the task.

        This method issues a request to the API.
        """
        # If the task is in a terminal status and we already have the status,
        # return it without refreshing it from the API.
        # Can't call is_terminal because it calls get status (infinite loop)
        if (self._status is not None and
                self._status in self.TERMINAL_STATUSES):
            return self._status

        resp = self._api.get_task_status(self._get_path_params())

        return models.TaskStatusCode(resp.body["status"])

    def get_info(self) -> Dict[str, Any]:
        """Get a dictionary with information about the task.

        Information includes e.g., "task_id", "status", timestamps
        ("create_time", "input_submit_time, "start_time", "end_time"),
        among others.

        This method issues a request to the API.
        """
        # If the task is in a terminal status and we already have the info,
        # return it without refreshing it from the API.
        if self._info is not None and self.is_terminal():
            return self._info

        params = self._get_path_params()
        resp = self._api.get_task(params, skip_deserialization=True).response

        info = json.loads(resp.data.decode("utf-8"))
        status = models.TaskStatusCode(info["status"])

        self._info = info
        self._status = status

        return info

    def wait(self, polling_period: int = 5) -> models.TaskStatusCode:
        """Wait for the task to complete.

        This method issues requests to the API.

        Args:
            polling_period: How often to poll the API for the task status.

        Returns:
            The final status of the task.
        """
        prev_status = None
        while True:
            status = self.get_status()
            if status != prev_status:
                if status == models.TaskStatusCode.PENDINGINPUT:
                    pass
                elif status == models.TaskStatusCode.SUBMITTED:
                    logging.info(
                        "Task %s successfully queued and waiting to be "
                        "picked-up for execution...", self.id)
                elif status == models.TaskStatusCode.STARTED:
                    logging.info(
                        "Task %s has started and is now running "
                        "remotely.", self.id)
                elif status == models.TaskStatusCode.SUCCESS:
                    logging.info("Task %s completed successfully.", self.id)
                elif status == models.TaskStatusCode.FAILED:
                    logging.info("Task %s failed.", self.id)
                    logging.info("Download the 'stdout.txt' and 'stderr.txt' "
                                 "files with `task.download_outputs()` for "
                                 "more detail.")
                elif status == models.TaskStatusCode.PENDINGKILL:
                    logging.info("Task %s is being killed.", self.id)
                elif status == models.TaskStatusCode.KILLED:
                    logging.info("Task %s killed.", self.id)
                elif status == models.TaskStatusCode.ZOMBIE:
                    logging.info("The machine was terminated while the task "
                                 "was pending.")
                else:
                    logging.info(
                        "An internal error occurred with status %s "
                        "while performing the task.", status)
            prev_status = status

            if self.is_terminal():
                return status

            time.sleep(polling_period)

    def _send_kill_request(self, max_api_requests: int) -> None:
        """Send a kill request to the API.
        If the api request fails, it will retry until
        max_api_requests is 0 raising a RuntimeError.
        Args:
            max_api_requests (int): maximum number of api requests to send
        """
        while max_api_requests > 0:
            max_api_requests -= 1
            try:
                if self.is_terminal():
                    break

                path_params = self._get_path_params()
                self._api.kill_task(path_params=path_params)
                break
            except exceptions.ApiException as exc:
                if max_api_requests == 0:
                    raise RuntimeError(
                        "Something went wrong while sending"
                        " the kill command. Please try again later.") from exc
                time.sleep(constants.TASK_KILL_RETRY_SLEEP_SEC)

    def _check_if_pending_kill(
            self,
            wait_timeout: Union[float,
                                int]) -> Tuple[bool, models.TaskStatusCode]:
        """Check if the task is in the PENDINGKILL state.
        This method keeps checking the status of the task until it is no longer
        in the PENDINGKILL state or until the timeout is reached.
        Args:
            wait_timeout (int, float): number of seconds to wait for the
            state to leave PENDINGKILL.
        Returns:
            A tuple with a boolean indicating whether the timeout was reached
            and the status of the task.
        """
        success = True
        start = time.time()

        while (status :=
               self.get_status()) == models.TaskStatusCode.PENDINGKILL:
            if (time.time() - start) > wait_timeout:
                success = False
                break

            time.sleep(constants.TASK_KILL_RETRY_SLEEP_SEC)
        return success, status

    def kill(self,
             wait_timeout: Optional[Union[float, int]] = None,
             verbosity_level: int = 2) -> Union[bool, None]:
        """Request a task to be killed.
        
        This method requests that the current task is remotely killed.
        If `wait_timeout` is None (default), the kill request is sent to the
        backend and the method returns. However, if `wait_timeout` is 
        a positive number, the method waits up to `wait_timeout` seconds
        to ensure that the task transitions to the KILLED state.
        Args:
            wait_timeout (int, float): Optional - number of seconds to wait
            for the kill command or None if only the request is to be sent.
            verbosity_level (int): Optional. the verbosity of the logs when the
            task signal is sent and when the task is killed. Verbosity 0
            produces no outputs, 1 produces minimal outputs, and 2 (Default)
            produces extensive outputs.
        Returns:
            - None if `wait_timeout` is None and the kill request was
              successfully sent;
            - True if `wait_timeout`> 0 and the task successfully transitioned
              to the `KILLED` state within `wait_timeout` seconds;
            - False if `wait_timeout` > 0 but the task didn't transition
              to the `KILLED` state within `wait_timeout` seconds;

        """
        if wait_timeout is not None:
            if not isinstance(wait_timeout, (float, int)):
                raise TypeError("Wait timeout must be a number.")
            if wait_timeout <= 0.0:
                raise ValueError("Wait timeout must be a positive number.")

        if verbosity_level not in self.KILL_VERBOSITY_LEVELS:
            raise ValueError(f"Verbosity {verbosity_level} level not allowed. "
                             f"Choose from {self.KILL_VERBOSITY_LEVELS}")

        self._send_kill_request(constants.TASK_KILL_MAX_API_REQUESTS)

        if wait_timeout is None:
            logging.info(
                __("task-kill-request-sent" + f"-{verbosity_level}", self.id))
            return None

        success, status = self._check_if_pending_kill(wait_timeout)

        if status != models.TaskStatusCode.KILLED:
            success = False
            if status == models.TaskStatusCode.PENDINGKILL:
                logging.error(
                    "Unable to ensure that task %s transitioned to the KILLED "
                    "state after %f seconds. The status of the task is %s.",
                    self.id, wait_timeout, status)
            else:
                logging.error(
                    "Task is already in a terminal state and cannot be killed. "
                    "Current task status is %s.", status)

        if success:
            if verbosity_level == 2:
                logging.info("Successfully killed task %s.", self.id)
            elif verbosity_level == 1:
                logging.info("%s killed.")

        return success

    def get_simulator_name(self) -> str:
        # e.g. retrieve openfoam from fvm.openfoam.run_simulation
        return self.get_info()["method_name"].split(".")[1]

    def get_storage_path(self) -> str:
        return self.get_info()["storage_path"]

    def get_output_files_info(self) -> output_contents.OutputContents:
        """Get information of the output files of the task.

        Returns:
            An instance of the OutputContents class, which can be used to
            access info about the output files, such as the archive's size,
            number of files, and information about each file (name, size,
            compressed size). It can also be used to print that information
            in a formatted way.
        """
        api_response = self._api.get_outputs_list(
            path_params=self._get_path_params())

        archive_info = api_response.body

        contents = [
            output_contents.FileInfo(
                name=file_info["name"],
                size=file_info["size"],
                compressed_size=file_info["compressed_size"],
            ) for file_info in archive_info["contents"]
        ]

        return output_contents.OutputContents(
            size=int(archive_info["size"]),
            contents=contents,
        )

    def download_outputs(
        self,
        filenames: Optional[List[str]] = None,
        output_dir: Optional[types.Path] = None,
        uncompress: bool = True,
        rm_downloaded_zip_archive: bool = True,
    ) -> pathlib.Path:
        """Download output files of the task.

        Args:
            filenames: List of filenames to download. If None or empty, all
                files are downloaded.
            output_dir: Directory where to download the files. If None, the
                files are downloaded to the default directory. The default is
                {inductiva.get_output_dir()}/{output_dir}/{task_id}.
            uncompress: Whether to uncompress the archive after downloading it.
            rm_downloaded_zip_archive: Whether to remove the archive after
            uncompressing it. If uncompress is False, this argument is ignored.
        """
        api_response = self._api.download_task_output(
            path_params=self._get_path_params(),
            query_params={
                "filename": filenames or [],
            },
            stream=True,
            skip_deserialization=True,
        )
        # use raw urllib3 response instead of the generated client response, to
        # implement our own download logic (with progress bar, first checking
        # the size of the file, etc.)
        response = api_response.response

        if output_dir is None:
            output_dir = files.resolve_output_path(self.id)
        else:
            output_dir = files.resolve_output_path(output_dir)

        if output_dir.exists():
            warnings.warn("Path already exists, files may be overwritten.")
        output_dir.mkdir(parents=True, exist_ok=True)

        zip_path = output_dir.joinpath("output.zip")

        logging.info("Downloading simulation outputs to %s.", zip_path)
        data.download_file(response, zip_path)

        if uncompress:
            logging.info("Uncompressing the outputs to %s.", output_dir)
            data.uncompress_task_outputs(zip_path, output_dir)
            if rm_downloaded_zip_archive:
                zip_path.unlink()
        return output_dir

    class _PathParams(TypedDict):
        """Util class for type checking path params."""
        task_id: str

    def _get_path_params(self) -> _PathParams:
        """Get dictionary with the URL path parameters for API calls."""
        return {"task_id": self.id}

    def get_computation_time(self,
                             fail_if_running: bool = True) -> Optional[float]:
        """Get the time the computation of the task took to complete.

        Returns:
            The time in hh mm ss or None if the task hasn't completed yet.
        """
        info = self.get_info()
        if fail_if_running and not self.is_terminal():
            return None
        # start_time may be None if the task was killed before it started
        if info["computation_start_time"] is None:
            return None

        # Format the time to datetime type
        start_time = datetime.datetime.fromisoformat(
            info["computation_start_time"])
        end_time = info.get("computation_end_time")
        if end_time is None:
            end_time = datetime.datetime.now(datetime.timezone.utc)
        else:
            end_time = datetime.datetime.fromisoformat(
                info["computation_end_time"])

        total_seconds = (end_time - start_time).total_seconds()
        return format_utils.seconds_formatter(total_seconds)

    def get_total_time(self, fail_if_running: bool = True) -> Optional[float]:
        """Get the total time the task workflow took to complete.

        Returns:
            The time in hh mm ss or None if the task hasn't completed yet.
        """
        info = self.get_info()
        if fail_if_running and not self.is_terminal():
            return None
        # start_time may be None if the task was killed before it started
        if info["input_submit_time"] is None:
            return None

        # Format the time to datetime type
        submitted_time = datetime.datetime.fromisoformat(
            info["input_submit_time"])
        end_time = info.get("end_time")

        if end_time is None:
            end_time = datetime.datetime.now(datetime.timezone.utc)
        else:
            end_time = datetime.datetime.fromisoformat(info["end_time"])

        total_seconds = (end_time - submitted_time).total_seconds()
        return format_utils.seconds_formatter(total_seconds)

    def get_machine_type(self) -> Optional[str]:
        """Get the machine type used in the task.

        Streamlines the process of obtaining the task info, extracting the
        machine type from the comprehensive task info.

        Returns:
            The machine type, or None if a machine hasn't been assigned yet.
        """
        info = self.get_info()
        if info["executer"] is None:
            return None

        machine_info = info["executer"]
        if "vm_type" not in machine_info:
            return "inductiva-machine"

        machine_type = machine_info["vm_type"].split("/")[-1]

        return machine_type



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class ExecuterTrackerRegisterInfo(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Info for creating an executer.
    """


    class MetaOapg:
        required = {
            "memory",
            "create_time",
            "cpu_count_logical",
            "cpu_count_physical",
            "git_commit_hash",
            "host_info",
            "supported_executer_types",
        }
        
        class properties:
            create_time = schemas.DateTimeSchema
            
            
            class supported_executer_types(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    items = schemas.StrSchema
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, str, ]], typing.List[typing.Union[MetaOapg.items, str, ]]],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'supported_executer_types':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)
            cpu_count_logical = schemas.IntSchema
            cpu_count_physical = schemas.IntSchema
            memory = schemas.IntSchema
            git_commit_hash = schemas.StrSchema
            
            
            class host_info(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    
                    @classmethod
                    @functools.lru_cache()
                    def one_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            GCloudHostInfo,
                            InductivaHostInfo,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'host_info':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            machine_group_id = schemas.UUIDSchema
            mpi_cluster = schemas.BoolSchema
            num_mpi_hosts = schemas.IntSchema
            __annotations__ = {
                "create_time": create_time,
                "supported_executer_types": supported_executer_types,
                "cpu_count_logical": cpu_count_logical,
                "cpu_count_physical": cpu_count_physical,
                "memory": memory,
                "git_commit_hash": git_commit_hash,
                "host_info": host_info,
                "machine_group_id": machine_group_id,
                "mpi_cluster": mpi_cluster,
                "num_mpi_hosts": num_mpi_hosts,
            }
    
    memory: MetaOapg.properties.memory
    create_time: MetaOapg.properties.create_time
    cpu_count_logical: MetaOapg.properties.cpu_count_logical
    cpu_count_physical: MetaOapg.properties.cpu_count_physical
    git_commit_hash: MetaOapg.properties.git_commit_hash
    host_info: MetaOapg.properties.host_info
    supported_executer_types: MetaOapg.properties.supported_executer_types
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["create_time"]) -> MetaOapg.properties.create_time: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["supported_executer_types"]) -> MetaOapg.properties.supported_executer_types: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["cpu_count_logical"]) -> MetaOapg.properties.cpu_count_logical: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["cpu_count_physical"]) -> MetaOapg.properties.cpu_count_physical: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["memory"]) -> MetaOapg.properties.memory: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["git_commit_hash"]) -> MetaOapg.properties.git_commit_hash: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["host_info"]) -> MetaOapg.properties.host_info: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["machine_group_id"]) -> MetaOapg.properties.machine_group_id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["mpi_cluster"]) -> MetaOapg.properties.mpi_cluster: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["num_mpi_hosts"]) -> MetaOapg.properties.num_mpi_hosts: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["create_time", "supported_executer_types", "cpu_count_logical", "cpu_count_physical", "memory", "git_commit_hash", "host_info", "machine_group_id", "mpi_cluster", "num_mpi_hosts", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["create_time"]) -> MetaOapg.properties.create_time: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["supported_executer_types"]) -> MetaOapg.properties.supported_executer_types: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["cpu_count_logical"]) -> MetaOapg.properties.cpu_count_logical: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["cpu_count_physical"]) -> MetaOapg.properties.cpu_count_physical: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["memory"]) -> MetaOapg.properties.memory: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["git_commit_hash"]) -> MetaOapg.properties.git_commit_hash: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["host_info"]) -> MetaOapg.properties.host_info: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["machine_group_id"]) -> typing.Union[MetaOapg.properties.machine_group_id, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["mpi_cluster"]) -> typing.Union[MetaOapg.properties.mpi_cluster, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["num_mpi_hosts"]) -> typing.Union[MetaOapg.properties.num_mpi_hosts, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["create_time", "supported_executer_types", "cpu_count_logical", "cpu_count_physical", "memory", "git_commit_hash", "host_info", "machine_group_id", "mpi_cluster", "num_mpi_hosts", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        memory: typing.Union[MetaOapg.properties.memory, decimal.Decimal, int, ],
        create_time: typing.Union[MetaOapg.properties.create_time, str, datetime, ],
        cpu_count_logical: typing.Union[MetaOapg.properties.cpu_count_logical, decimal.Decimal, int, ],
        cpu_count_physical: typing.Union[MetaOapg.properties.cpu_count_physical, decimal.Decimal, int, ],
        git_commit_hash: typing.Union[MetaOapg.properties.git_commit_hash, str, ],
        host_info: typing.Union[MetaOapg.properties.host_info, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
        supported_executer_types: typing.Union[MetaOapg.properties.supported_executer_types, list, tuple, ],
        machine_group_id: typing.Union[MetaOapg.properties.machine_group_id, str, uuid.UUID, schemas.Unset] = schemas.unset,
        mpi_cluster: typing.Union[MetaOapg.properties.mpi_cluster, bool, schemas.Unset] = schemas.unset,
        num_mpi_hosts: typing.Union[MetaOapg.properties.num_mpi_hosts, decimal.Decimal, int, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'ExecuterTrackerRegisterInfo':
        return super().__new__(
            cls,
            *_args,
            memory=memory,
            create_time=create_time,
            cpu_count_logical=cpu_count_logical,
            cpu_count_physical=cpu_count_physical,
            git_commit_hash=git_commit_hash,
            host_info=host_info,
            supported_executer_types=supported_executer_types,
            machine_group_id=machine_group_id,
            mpi_cluster=mpi_cluster,
            num_mpi_hosts=num_mpi_hosts,
            _configuration=_configuration,
            **kwargs,
        )

from inductiva.client.model.g_cloud_host_info import GCloudHostInfo
from inductiva.client.model.inductiva_host_info import InductivaHostInfo



"""GROMACS module of the API"""

from typing import Optional

from inductiva import types, tasks, simulators


class GROMACS(simulators.Simulator):
    """Class to invoke any GROMACS command on the API."""

    def __init__(self):
        super().__init__()
        self.api_method_name = "md.gromacs.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        commands: types.Commands,
        on: Optional[types.ComputationalResources] = None,
        storage_dir: Optional[types.Path] = "",
        extra_metadata: Optional[dict] = None,
    ) -> tasks.Task:
        """Run a list of GROMACS commands.

        Args:
            input_dir: Path to the directory containing the input files.
            commands: List of commands to run using the GROMACS simulator.
            on: The computational resource to launch the simulation on. If None
                the simulation is submitted to a machine in the default pool.
            storage_dir: Parent directory for storing simulation
                               results.
        """

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           extra_metadata=extra_metadata)



"""FEniCSx module of the API for Finite Element Analysis."""

from typing import Optional

from inductiva import simulators, types, tasks


class FEniCSx(simulators.Simulator):
    """Class to invoke a generic FEniCSx simulation on the API."""

    def __init__(self):
        super().__init__()
        self.api_method_name = "fem.fenicsx.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        geometry_filename: str,
        bcs_filename: str,
        material_filename: str,
        global_refinement_meshing_factor: float = 1.0,
        local_refinement_meshing_factor: float = 1.0,
        smoothing_meshing_parameter: float = 10.0,
        mesh_element_family: str = "CG",
        mesh_element_order: int = 1,
        on: Optional[types.ComputationalResources] = None,
        storage_dir: Optional[types.Path] = "",
        extra_metadata: Optional[dict] = None,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            geometry_filename: Geometry filename.
            bcs_filename: Boundary conditions filename.
            material_filename: Material filename.
            global_refinement_meshing_factor (float): The refinement factor for
              global refinement of the mesh. A higher value results in a finer
              mesh overall, increasing the number of elements in the entire
              mesh, and leading to a more detailed representation of the
              geometry. Use this factor when you want to globally refine the
              mesh uniformly, without specific local focus.
            local_refinement_meshing_factor (float): The refinement factor for
              local refinement of the mesh. This factor controls the local
              refinement level of the mesh and is typically used for refining
              specific regions or features of the mesh. A higher value for this
              factor indicates a finer mesh in the regions of interest,
              providing more detailed resolution around certain features. Use
              this factor when you want to focus on refining specific areas
              while keeping the rest of the mesh less refined.
            smoothing_parameter (float): The smoothing parameter for mesh
              generation. It controls the amount of mesh smoothing applied to
              the generated mesh. Adjust this parameter for improved mesh
              quality.
            mesh_element_family (str): The type of mesh element family.
            mesh_element_order (int): The (polynomial) order of the mesh
              element.
            on: The computational resource to launch the simulation on. If None
                the simulation is submitted to a machine in the default pool.
            storage_dir: Parent directory for storing simulation results.
            other arguments: See the documentation of the base class.
        """

        return super().run(
            input_dir,
            on=on,
            geometry_filename=geometry_filename,
            bcs_filename=bcs_filename,
            material_filename=material_filename,
            global_refinement_meshing_factor=global_refinement_meshing_factor,
            local_refinement_meshing_factor=local_refinement_meshing_factor,
            smoothing_meshing_parameter=smoothing_meshing_parameter,
            mesh_element_family=mesh_element_family,
            mesh_element_order=mesh_element_order,
            storage_dir=storage_dir,
            extra_metadata=extra_metadata,
        )



"""CLI commands for listing both the active and available machine groups."""

import argparse

from inductiva import resources


def print_cpu_series_info(cpu_series: str,
                          cpu_series_info: dict,
                          verbose: bool = False):
    """Format and print the information for given CPU series."""

    cpu_series_description = cpu_series_info["description"]
    print(f"{cpu_series}: {cpu_series_description}")
    for ram_type, type_info in cpu_series_info["types"].items():
        description = ("-> " + type_info["description"]) if verbose else ""
        vcpus = str(type_info["vcpus"])
        cpu_type = f"{cpu_series}-{ram_type}-"
        print(f"  > {cpu_type:13} {vcpus:45} {description}")
    print()


def list_machine_types_available(args):
    """List available machine types information per provider and CPU series."""

    provider = args.provider
    description = args.description
    cpu_series = args.series

    resources_available = resources.machine_types.get_available()
    provider_resources = resources_available[provider]
    print(provider_resources["description"] + "\n")

    if cpu_series is not None:
        cpu_series_info = provider_resources["cpu-series"][cpu_series]
        print_cpu_series_info(cpu_series, cpu_series_info, description)
    else:
        # Print all available machine CPU series information
        for cpu_series, series_info in provider_resources["cpu-series"].items():
            print_cpu_series_info(cpu_series, series_info, description)


def list_machine_groups(unused_args):
    """List Resources."""
    resources.machine_groups.list()


def register(parser):
    """Register the list resources commands."""

    subparser = parser.add_parser("available",
                                  help="List available machine types.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.add_argument("-p",
                           "--provider",
                           type=str,
                           default="gcp",
                           choices=["gcp"],
                           help="Filter the available types by provider.")
    subparser.add_argument("-s",
                           "--series",
                           default=None,
                           type=str,
                           help="Filter the available types by CPU series.")
    subparser.add_argument("-d",
                           "--description",
                           action="store_true",
                           help="Show a description of the machine types.")

    subparser.description = (
        "The `inductiva available` command provides a utility for listing the\n"
        "available machine types by provider (default: gcp) and CPU series.\n"
        "The list includes a description of the memory types and vCPUs "
        "available.\n\n")

    subparser.set_defaults(func=list_machine_types_available)

    subparser = parser.add_parser("list",
                                  aliases=["ls"],
                                  help="List currently active resources.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva resources list` command provides a snapshot "
        "of your active computational resources.\nFrom machine type "
        "to start time, it gives you a comprehensive overview of "
        "your resources in one place.")
    subparser.set_defaults(func=list_machine_groups)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class MachineTypeResponse(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "machine_type",
            "num_cpus",
            "price",
            "ram_gb",
        }
        
        class properties:
            machine_type = schemas.StrSchema
            num_cpus = schemas.IntSchema
            ram_gb = schemas.IntSchema
            price = schemas.NumberSchema
            
            
            class provider(
                schemas.EnumBase,
                schemas.StrSchema
            ):
                
                @schemas.classproperty
                def GCP(cls):
                    return cls("GCP")
            __annotations__ = {
                "machine_type": machine_type,
                "num_cpus": num_cpus,
                "ram_gb": ram_gb,
                "price": price,
                "provider": provider,
            }
    
    machine_type: MetaOapg.properties.machine_type
    num_cpus: MetaOapg.properties.num_cpus
    price: MetaOapg.properties.price
    ram_gb: MetaOapg.properties.ram_gb
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["machine_type"]) -> MetaOapg.properties.machine_type: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["num_cpus"]) -> MetaOapg.properties.num_cpus: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["ram_gb"]) -> MetaOapg.properties.ram_gb: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["price"]) -> MetaOapg.properties.price: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["provider"]) -> MetaOapg.properties.provider: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["machine_type", "num_cpus", "ram_gb", "price", "provider", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["machine_type"]) -> MetaOapg.properties.machine_type: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["num_cpus"]) -> MetaOapg.properties.num_cpus: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["ram_gb"]) -> MetaOapg.properties.ram_gb: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["price"]) -> MetaOapg.properties.price: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["provider"]) -> typing.Union[MetaOapg.properties.provider, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["machine_type", "num_cpus", "ram_gb", "price", "provider", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        machine_type: typing.Union[MetaOapg.properties.machine_type, str, ],
        num_cpus: typing.Union[MetaOapg.properties.num_cpus, decimal.Decimal, int, ],
        price: typing.Union[MetaOapg.properties.price, decimal.Decimal, int, float, ],
        ram_gb: typing.Union[MetaOapg.properties.ram_gb, decimal.Decimal, int, ],
        provider: typing.Union[MetaOapg.properties.provider, str, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'MachineTypeResponse':
        return super().__new__(
            cls,
            *_args,
            machine_type=machine_type,
            num_cpus=num_cpus,
            price=price,
            ram_gb=ram_gb,
            provider=provider,
            _configuration=_configuration,
            **kwargs,
        )



"""Functions to manage or retrieve user resources."""
from typing import Optional
from absl import logging
import inductiva
import inductiva.client
from inductiva.client.apis.tags import compute_api
from inductiva.utils import format_utils
from inductiva import resources
from . import machines_base


def estimate_machine_cost(machine_type: str, spot: bool = False):
    """Estimate the cloud cost of one machine per hour in US dollars.

    Args:
        machine_type: The type of GC machine to launch. Ex: "e2-standard-4".
            Check https://cloud.google.com/compute/docs/machine-resource for
            more information about machine types.
        zone: The zone where the machines will be launched.
        spot: Whether to use spot machines.
    """

    api = compute_api.ComputeApi(inductiva.api.get_client())

    instance_price = api.get_instance_price({
        "machine_type": machine_type,
    })

    if spot:
        estimated_cost = instance_price.body["preemptible_price"]
    else:
        estimated_cost = instance_price.body["on_demand_price"]

    return float(estimated_cost)


def _machine_group_list_to_str(machine_group_list) -> str:
    """Returns a string representation of a list of machine groups."""
    columns = [
        "Name",
        "Machine Type",
        "Elastic",
        "Type",
        "# machines",
        "Data Size in GB",
        "Spot",
        "Started at (UTC)",
    ]
    rows = []

    for machine_group in machine_group_list:
        is_elastic = False
        resource_type = machines_base.ResourceType.STANDARD.value
        spot = machine_group.spot if hasattr(machine_group, "spot") else False

        if isinstance(machine_group, resources.ElasticMachineGroup):
            is_elastic = True
        else:
            if isinstance(machine_group, resources.MPICluster):
                resource_type = machines_base.ResourceType.MPI.value
        num_active_machines = machine_group.active_machines_to_str()
        rows.append([
            machine_group.name, machine_group.machine_type, is_elastic,
            resource_type, num_active_machines, machine_group.data_disk_gb,
            spot, machine_group.create_time
        ])

    formatters = {
        "Started at (UTC)": [format_utils.datetime_formatter],
    }

    emph_formatter = format_utils.get_ansi_formatter()
    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    return format_utils.get_tabular_str(rows,
                                        columns,
                                        formatters=formatters,
                                        header_formatters=header_formatters)


def _fetch_machine_groups_from_api():
    """Get all active machine groups of a user from the API."""
    try:
        api = compute_api.ComputeApi(inductiva.api.get_client())
        response = api.list_active_user_instance_groups()
        if len(response.body) == 0:
            print("No active computational resources found.")
            return response.body

        return response.body

    except inductiva.client.ApiException as api_exception:
        raise api_exception


# pylint: disable=redefined-builtin
def list():
    # pylint: disable=line-too-long
    """Lists all active resources info.

    This method queries Google Cloud for active resources
    that belong to the user. It outputs all the information relative to
    each resource as folllows:

    Active Resources:
                                        Name         VM Type    Elastic   # machines    Disk Size in GB       Spot   Started at (UTC)
    api-6359c03d-c4f9-479f-8b11-ba1f8f55a58c   e2-standard-4      False            3                 40      False   10 Oct, 13:40:50
    api-db2046cf-a6fc-4124-926c-1a24329da5ea   e2-standard-4       True          2/4                 40      False   10 Oct, 12:43:03
    """
    # pylint: enable=line-too-long

    machine_group_list = get()
    if len(machine_group_list) != 0:
        print("Active Resources:")
        print(_machine_group_list_to_str(machine_group_list))


def get():
    """Returns a list of 'Resource' objects."""

    # Retrive the active resource names
    machine_groups = _fetch_machine_groups_from_api()
    machine_group_list = []

    for mg in machine_groups:
        if mg["is_elastic"]:
            mg_class = resources.ElasticMachineGroup
        elif mg["type"] == "standard":
            mg_class = resources.MachineGroup
        elif mg["type"] == "mpi":
            mg_class = resources.MPICluster
        else:
            raise ValueError("Unknown resource configuration.")
        machine_group_list.append(mg_class.from_api_response(mg))

    return machine_group_list


def get_cheapest_machine_type(num_cpus: int,
                              ram_gb: Optional[int] = None,
                              spot: bool = False):
    """Get the machine type with the lowest price.

    If the machine type with the exact requirements is not found, the closest
    machine type with higher resources will be returned. For example, if a
    machine type with 5 CPUs and 20 GB of RAM is requested, but the closest
    machine type has 8 CPUs and 32 GB of RAM, the latter will be returned.

    Args:
        num_cpus: The minimum number of CPUs required.
        ram_gb: The minimum RAM in GB required. If None, the minimum RAM for
          the given number of CPUs will be used.
        spot: Whether the machine is spot or not.
    """
    spot = "t" if spot else "f"
    api = compute_api.ComputeApi(inductiva.api.get_client())
    body = {"num_cpus": num_cpus, "spot": spot}
    if ram_gb:
        body["ram_gb"] = ram_gb
    response = api.get_machine_type(body)

    logging.info("Machine type: %s", response.body["machine_type"])
    logging.info("CPUs: %s", response.body["num_cpus"])
    logging.info("RAM in GB: %s", response.body["ram_gb"])
    logging.info("Estimated cost per hour: %s $/h",
                 round(response.body["price"], 5))

    return response.body["machine_type"]



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class GCloudHostInfo(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Info about Google Cloud machine hosting the executer.
    """


    class MetaOapg:
        required = {
            "vm_id",
            "vm_type",
            "preemptible",
            "vm_zone",
            "host_type",
            "vm_name",
        }
        
        class properties:
            
            
            class host_type(
                schemas.EnumBase,
                schemas.StrSchema
            ):
                
                @schemas.classproperty
                def GCLOUD(cls):
                    return cls("gcloud")
            vm_type = schemas.StrSchema
            vm_name = schemas.StrSchema
            vm_id = schemas.StrSchema
            vm_zone = schemas.StrSchema
            preemptible = schemas.BoolSchema
            __annotations__ = {
                "host_type": host_type,
                "vm_type": vm_type,
                "vm_name": vm_name,
                "vm_id": vm_id,
                "vm_zone": vm_zone,
                "preemptible": preemptible,
            }
    
    vm_id: MetaOapg.properties.vm_id
    vm_type: MetaOapg.properties.vm_type
    preemptible: MetaOapg.properties.preemptible
    vm_zone: MetaOapg.properties.vm_zone
    host_type: MetaOapg.properties.host_type
    vm_name: MetaOapg.properties.vm_name
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["host_type"]) -> MetaOapg.properties.host_type: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["vm_type"]) -> MetaOapg.properties.vm_type: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["vm_name"]) -> MetaOapg.properties.vm_name: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["vm_id"]) -> MetaOapg.properties.vm_id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["vm_zone"]) -> MetaOapg.properties.vm_zone: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["preemptible"]) -> MetaOapg.properties.preemptible: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["host_type", "vm_type", "vm_name", "vm_id", "vm_zone", "preemptible", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["host_type"]) -> MetaOapg.properties.host_type: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["vm_type"]) -> MetaOapg.properties.vm_type: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["vm_name"]) -> MetaOapg.properties.vm_name: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["vm_id"]) -> MetaOapg.properties.vm_id: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["vm_zone"]) -> MetaOapg.properties.vm_zone: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["preemptible"]) -> MetaOapg.properties.preemptible: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["host_type", "vm_type", "vm_name", "vm_id", "vm_zone", "preemptible", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        vm_id: typing.Union[MetaOapg.properties.vm_id, str, ],
        vm_type: typing.Union[MetaOapg.properties.vm_type, str, ],
        preemptible: typing.Union[MetaOapg.properties.preemptible, bool, ],
        vm_zone: typing.Union[MetaOapg.properties.vm_zone, str, ],
        host_type: typing.Union[MetaOapg.properties.host_type, str, ],
        vm_name: typing.Union[MetaOapg.properties.vm_name, str, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'GCloudHostInfo':
        return super().__new__(
            cls,
            *_args,
            vm_id=vm_id,
            vm_type=vm_type,
            preemptible=preemptible,
            vm_zone=vm_zone,
            host_type=host_type,
            vm_name=vm_name,
            _configuration=_configuration,
            **kwargs,
        )



"""Mixin for file management.

Provides utility classes and functions for managing files and directories
and rendering template files for generation of simulation inputs.
"""
from typing import Union
import pathlib
import shutil
import glob
import io
import os
import re

from absl import logging
import jinja2

from inductiva.utils import format_utils

TEMPLATE_EXTENSION = ".jinja"
SUFFIX_SEPARATOR = "__"


class FileManager:
    """Class for file management."""
    DEFAULT_ROOT_DIR = "input_dir"
    __root_dir = None  #pylint: disable=invalid-name

    def __check_root_dir(self):
        """Set default root folder if unset."""
        if self.__root_dir is None:
            self.set_root_dir(self.DEFAULT_ROOT_DIR)

    def set_root_dir(self, root_dir=None):
        """Set a root directory for the file manager.
        
        All files managed through the manager will be
        placed inside a newly created folder with the given
        name. If a folder with the same name already exists,
        the newly created one will be appended with the
        "__1" suffix. If folders having the "__N" suffix already
        exist, the given name will be appended with "__{N+1}"
        where N is determined by the directory with the largest N.

        Args:
            root_dir: Path to the root directory.
                If None, an error is raised.
        """

        if root_dir is None:
            raise ValueError("Given root directory cannot be None")

        if os.path.isdir(root_dir):
            if format_utils.getenv_bool(
                    "INDUCTIVA_DISABLE_FILEMANAGER_AUTOSUFFIX", False):
                raise FileExistsError(f"Directory {root_dir} already exists.")
            generated_root_dir = _gen_unique_name(str(root_dir))
            logging.info(
                "Directory %s already exists."
                " Setting root folder to %s.", root_dir, generated_root_dir)
            root_dir = generated_root_dir

        logging.info("Setting root folder to %s.", root_dir)
        os.makedirs(root_dir)
        self.__root_dir = pathlib.Path(root_dir).resolve()

    def get_root_dir(self):
        """Get the active root directory for the file manager."""

        self.__check_root_dir()

        return self.__root_dir

    def add_file(self, source_file, target_file=None, **render_args):
        """Add a file to the root_dir and render in case it is a template file. 

        This method copies the contents of the `source_file` to a `target_file`
        inside the root directory. If the source file is a template file, it is
        rendered using the given `render_args` before copying to the target
        destination.

        If `target_file` is None, the copied file is saved inside the manager's
        root directory with the same name as the source file; if it is a
        template file, the template suffix is removed. 
        A file is considered to be a template file if it ends with the `.jinja`
        suffix. The rendering arguments are ignored if the source file is not a
        template file.

        Args:
            source_file: Path to the source file.
            target_file: Path to the target file.
            render_args: Arguments to render the template file.

        Examples:
        a) Add a file to the root directory "example_a"
            (input_file.txt -> example_a/input_file.txt):
            >>> file_manager = FileManager()
            >>> file_manager.set_root_dir("example_a")
            >>> file_manager.add_file("input_file.txt")
        
        b) Add a file to the root directory "example_b" with a different name
            (input_file.txt -> example_b/renamed_file.txt):
            >>> file_manager = FileManager()
            >>> file_manager.set_root_dir("example_b")
            >>> file_manager.add_file("input_file.txt", "renamed_file.txt")
        
        c) Render a template file to the root directory "example_c" with a
            template argument named "value" (input_file.txt.jinja ->
            example_c/input_file.txt):
            >>> file_manager = FileManager()
            >>> file_manager.set_root_dir("example_c")
            >>> file_manager.add_file("input_file.txt.jinja", value=2)
        """

        self.__check_root_dir()

        if target_file is None:
            new_target_file = pathlib.Path(source_file).name
        else:
            new_target_file = target_file

        new_target_file = os.path.join(self.__root_dir, new_target_file)
        if source_file.endswith(TEMPLATE_EXTENSION):
            if target_file is None:
                new_target_file, _ = os.path.splitext(new_target_file)

            render_file(template_file=source_file,
                        target_file=new_target_file,
                        **render_args)
        else:
            if render_args:
                logging.debug(
                    f"Ignoring render_args since %s"
                    f"isn't a {TEMPLATE_EXTENSION} file", source_file)
            shutil.copy(source_file, new_target_file)

        return new_target_file

    def add_dir(self, source_dir, target_dir=None, **render_args):
        """Add a directory to the root_dir and render all template files.

        This method copies the contents of the `source_dir` to a `target_dir`
        inside the root directory, keeping the structure as the source
        directory. All template files in the source directory are rendered
        using the given `render_args` before copying to the target destination.
        
        If `target_dir` is None, the source directory is copied to the root
        directory. All template files will have the template suffix stripped
        from their name upon copying.
        A file is considered to be a template file if it ends with the `.jinja`
        suffix. The rendering arguments are ignored no template file is
        available in the source directory.

        Args:
            source_dir: Path to the source directory.
            target_dir: Path to the target directory.
            render_args: Arguments to render the template files.

        Examples:
        a) Add a directory to the root directory "example_a"
            (input_dir/ -> example_a/input_dir/):
            >>> file_manager = FileManager()
            >>> file_manager.set_root_dir("example_a")
            >>> file_manager.add_dir("input_dir")

        b) Add directory to the root directory "example_b" with different name
            (input_dir/ -> example_b/renamed_dir/):
            >>> file_manager = FileManager()
            >>> file_manager.set_root_dir("example_b")
            >>> file_manager.add_dir("input_dir", "renamed_dir")
        
        c) Render template directory to the root directory with the template
        arguments named "value" and "name" (input_dir/ -> example_c/input_dir/):
            >>> file_manager = FileManager()
            >>> file_manager.set_root_dir("example_c")
            >>> file_manager.add_dir("input_dir", value=2, name="example")
        """
        self.__check_root_dir()

        if target_dir is None:
            target_dir = "."

        target_dir = os.path.join(self.__root_dir, target_dir)
        shutil.copytree(source_dir,
                        target_dir,
                        dirs_exist_ok=True,
                        symlinks=True)

        if render_args:
            template_files = get_template_files(target_dir)

            for template_file in template_files:
                template_path = os.path.join(target_dir, template_file)
                target_path = template_path.split(TEMPLATE_EXTENSION)[0]

                render_file(template_file=template_path,
                            target_file=target_path,
                            **render_args)
                os.remove(template_path)

        return target_dir


def render_file(template_file, target_file: Union[str, io.StringIO],
                **render_args):
    """Render a file from a template.

    Args:
        template_file: Path to the source file.
        target_file: Path to the target file.
        render_args: Arguments to render the template file.
    """

    source_path = pathlib.Path(template_file)

    source_dir = source_path.parent
    source_file = source_path.name

    environment = jinja2.Environment(loader=jinja2.FileSystemLoader(source_dir))
    template = environment.get_template(source_file)
    stream = template.stream(**render_args)
    stream.dump(target_file)


def get_template_files(src_dir):
    """Get all template files in a directory."""

    template_paths = glob.glob(os.path.join(src_dir, "**",
                                            "*" + TEMPLATE_EXTENSION),
                               recursive=True)

    template_files = [
        os.path.relpath(file_path, start=src_dir)
        for file_path in template_paths
    ]

    return template_files


def _gen_unique_suffix(name, filenames):
    """Generate a suffix for a filename, based on a list of files.
    
    Args:
        name: Name of the file.
        filenames: List of filenames.
    """
    regex = f"^{name}{SUFFIX_SEPARATOR}([0-9]+$)"

    max_suffix = -1
    for filename in filenames:
        if match := re.match(regex, filename):
            max_suffix = max(max_suffix, int(match.group(1)))

    exists_solo = name in filenames
    exists_derived = max_suffix >= 0

    if exists_derived:
        suffix = f"{SUFFIX_SEPARATOR}{max_suffix+1}"
    elif exists_solo:
        suffix = f"{SUFFIX_SEPARATOR}2"
    else:
        suffix = ""
    return suffix


def _gen_unique_name(name):
    """Generates an unique folder name.

    Generates an unique folder name derived from the given name and
    based on the names of the existing folders in the current working directory.
    If no folder exists with the same name as the given name,
    the input name will be returned; otherwise, a suffix is appended
    to the input name to make it unique. 

    See `_gen_unique_suffix` for more details.
    """

    name = name.strip()
    filenames = glob.glob(f"{name}*")
    return name + _gen_unique_suffix(name, filenames)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class ValidationError(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "msg",
            "loc",
            "type",
        }
        
        class properties:
            
            
            class loc(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    
                    
                    class items(
                        schemas.ComposedSchema,
                    ):
                    
                    
                        class MetaOapg:
                            any_of_0 = schemas.StrSchema
                            any_of_1 = schemas.IntSchema
                            
                            @classmethod
                            @functools.lru_cache()
                            def any_of(cls):
                                # we need this here to make our import statements work
                                # we must store _composed_schemas in here so the code is only run
                                # when we invoke this method. If we kept this at the class
                                # level we would get an error because the class level
                                # code would be run when this module is imported, and these composed
                                # classes don't exist yet because their module has not finished
                                # loading
                                return [
                                    cls.any_of_0,
                                    cls.any_of_1,
                                ]
                    
                    
                        def __new__(
                            cls,
                            *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                            _configuration: typing.Optional[schemas.Configuration] = None,
                            **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                        ) -> 'items':
                            return super().__new__(
                                cls,
                                *_args,
                                _configuration=_configuration,
                                **kwargs,
                            )
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ]], typing.List[typing.Union[MetaOapg.items, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ]]],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'loc':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)
            msg = schemas.StrSchema
            type = schemas.StrSchema
            __annotations__ = {
                "loc": loc,
                "msg": msg,
                "type": type,
            }
    
    msg: MetaOapg.properties.msg
    loc: MetaOapg.properties.loc
    type: MetaOapg.properties.type
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["loc"]) -> MetaOapg.properties.loc: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["msg"]) -> MetaOapg.properties.msg: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["type"]) -> MetaOapg.properties.type: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["loc", "msg", "type", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["loc"]) -> MetaOapg.properties.loc: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["msg"]) -> MetaOapg.properties.msg: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["type"]) -> MetaOapg.properties.type: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["loc", "msg", "type", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        msg: typing.Union[MetaOapg.properties.msg, str, ],
        loc: typing.Union[MetaOapg.properties.loc, list, tuple, ],
        type: typing.Union[MetaOapg.properties.type, str, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'ValidationError':
        return super().__new__(
            cls,
            *_args,
            msg=msg,
            loc=loc,
            type=type,
            _configuration=_configuration,
            **kwargs,
        )



"""Base class for low-level simulators."""
from typing import Optional
from abc import ABC

import pathlib

from inductiva import types, tasks, resources
from inductiva import commands


def mpi_enabled(cls):
    """Class decorator that adds MPICluster to the supported resources.
    """
    supports = getattr(cls, "_supported_resources", set())
    cls._supported_resources = supports | {resources.MPICluster}  # pylint: disable=protected-access

    return cls


class Simulator(ABC):
    """Base simulator class."""

    _supported_resources = {
        resources.MachineGroup, resources.ElasticMachineGroup
    }

    def __init__(self):
        self.api_method_name = ""

    @classmethod
    def get_supported_resources(cls):
        """Get the supported computational resources for this simulator."""
        return tuple(cls._supported_resources)

    def override_api_method_prefix(self, prefix: str):
        """Override the API method prefix.

        Example:
            # prefix = "protein_solvation"
            "md.gromacs.run_simulation" becomes
              "protein_solvation.gromacs.run_simulation"

        Args:
            prefix: The new prefix to use.
        """
        last_elements = self.api_method_name.split(".")[1:]
        all_elements = [prefix] + last_elements

        self.api_method_name = ".".join(all_elements)

    def _setup_input_dir(self, input_dir: types.Path):
        """Setup the simulator input directory."""
        input_dir = pathlib.Path(input_dir)
        if not input_dir.is_dir():
            raise ValueError(
                f"The provided path (\"{input_dir}\") is not a directory.")
        return input_dir

    def run(
        self,
        input_dir: types.Path,
        *_args,
        on: Optional[types.ComputationalResources] = None,
        storage_dir: Optional[types.Path] = "",
        extra_metadata: Optional[dict] = None,
        **kwargs,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory containing the input files.
            _args: Unused in this method, but defined to allow for more
                non-default arguments in method override in subclasses.
            on: The computational resource to launch the simulation in. If None
                the simulation is launched in a machine of the default pool.
            storage_dir: Parent directory for storing simulation
                               results.
            **kwargs: Additional keyword arguments to be passed to the
                simulation API method.
        """
        input_dir = self._setup_input_dir(input_dir)

        self.validate_computational_resources(on)

        if "commands" in kwargs:
            kwargs["commands"] = commands.Command.commands_to_dicts(
                kwargs["commands"])

        return tasks.run_simulation(
            self.api_method_name,
            input_dir,
            computational_resources=on,
            storage_dir=storage_dir,
            extra_metadata=extra_metadata,
            **kwargs,
        )

    def validate_computational_resources(self, resource):
        """Validate the computational resources passed to the run method.

        Args:
            resource: The computational resource to validate.
            valid_resources: The valid computational resources for the simulator
        """

        if resource is not None \
            and not isinstance(resource, tuple(self._supported_resources)):
            raise ValueError(
                "The computational resource is invalid for this simulator. "
                f"Expected one of {self._supported_resources} but got "
                f"{resource}.")



"""Classes to manage different Google Cloud machine group types."""
from absl import logging

from inductiva.resources import machines_base


class MachineGroup(machines_base.BaseMachineGroup):
    """Class to launch and manage a group of machines in Google Cloud.

    A machine group is a collection of homogenous machines with given the
    configurations that are launched in Google Cloud.
    Note: The machine group will be available only after calling 'start' method.
    The billing will start only after the machines are started."""

    def __init__(
        self,
        machine_type: str,
        num_machines: int = 1,
        spot: bool = False,
        data_disk_gb: int = 10,
        register: bool = True,
    ) -> None:
        """Create a MachineGroup object.

        The register argument is used to indicate if the machine group should
        be registered or if it was already registered. If set as False on
        initialization, then, the machine group is not registered and it
        can not be started in the cloud. This serves has an helper argument for
        retrieving already registered machine groups that can be started, for
        example, when retrieving with the `machines_groups.get` method.
        Users should not set this argument.

        Args:
            machine_type: The type of GC machine to launch. Ex: "e2-standard-4".
              Check https://cloud.google.com/compute/docs/machine-resource for
              information about machine types.
            num_machines: The number of virtual machines to launch.
            spot: Whether to use spot machines.
            data_disk_gb: The size of the disk for user data (in GB).
        """
        if num_machines < 1:
            raise ValueError(
                "`num_machines` should be a number greater than 0.")

        super().__init__(machine_type=machine_type,
                         data_disk_gb=data_disk_gb,
                         register=register)
        # Num_machines is the number of requested machines
        self.num_machines = num_machines
        #Number of active machines at the time of
        #the request machine_groups.get()
        self._active_machines = 0
        self.spot = spot
        self.__is_elastic = False

        if register:
            logging.info("Registering MachineGroup configurations:")
            self._register_machine_group(num_vms=self.num_machines,
                                         spot=self.spot,
                                         is_elastic=self.__is_elastic)

    @classmethod
    def from_api_response(cls, resp: dict):
        machine_group = super().from_api_response(resp)
        machine_group.num_machines = int(resp["max_vms"])
        machine_group.__dict__["_active_machines"] = int(resp["num_vms"])
        machine_group.spot = bool(resp["spot"])
        machine_group.register = False
        return machine_group

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}(name=\"{self.name}\")"

    def __str__(self):
        return f"Machine Group {self.name} with {self.machine_type} machines"

    def start(self):
        """Starts all machines of the machine group."""
        return super().start(num_vms=self.num_machines,
                             is_elastic=self.__is_elastic,
                             spot=self.spot)

    def terminate(self):
        """Terminates all machines of the machine group."""
        return super().terminate(num_vms=self.num_machines,
                                 is_elastic=self.__is_elastic,
                                 spot=self.spot)

    def _log_machine_group_info(self):
        super()._log_machine_group_info()
        logging.info("> Number of machines: %s", self.num_machines)
        logging.info("> Spot:               %s", self.spot)
        self.estimate_cloud_cost()

    def estimate_cloud_cost(self):
        """Estimates a cost per hour of the machine group in US dollars.

        This is only an estimate of having a machine group with the
        specified configurations up in the cloud. The actual cost may vary.

        Returns:
            The estimated cost per hour of the machine group, in US
              dollars ($/h)."""
        cost_per_machine = super()._get_estimated_cost(self.spot)
        estimated_cost = cost_per_machine * self.num_machines
        logging.info("> Estimated cloud cost of machine group: %.3f $/h",
                     estimated_cost)
        return estimated_cost


class ElasticMachineGroup(machines_base.BaseMachineGroup):
    """Manages an elastic machine group in Google Cloud.

    An ElasticMachineGroup is a set of identical machines that can
    automatically scale based on CPU load. The group starts with a
    minimum number of machines and adjusts its size as needed scaling
    to the maximum number of machines, ensuring both optimal performance
    and cost efficiency.

    Note: The machine group becomes active after calling the 'start' method,
    and billing commences once the machines are initiated.
    """

    def __init__(
        self,
        machine_type: str,
        min_machines: int = 1,
        max_machines: int = 1,
        spot: bool = False,
        data_disk_gb: int = 10,
        register: bool = True,
    ) -> None:
        """Create an ElasticMachineGroup object.

        The register argument is used to indicate if the machine group should
        be registered or if it was already registered. If set as False on
        initialization, then, the machine group is not registered and it
        can not be started in the cloud. This serves has an helper argument for
        retrieving already registered machine groups that can be started, for
        example, when retrieving with the `machines_groups.get` method.
        Users should not set this argument.
        
        Args:
            machine_type: The type of GC machine to launch. Ex: "e2-standard-4".
              Check https://cloud.google.com/compute/docs/machine-resource for
            more information about machine types.
            min_machines: The minimum number of available machines. This is
              a qunatity of machines that will be started initially and the
              minimum available machines, even in cases of low CPU load.
            max_machines: The maximum number of machines a machine group
              can scale up to.
            spot: Whether to use spot machines.
            data_disk_gb: The size of the disk for user data (in GB).
        """
        if min_machines < 1:
            raise ValueError(
                "`min_machines` should be a number greater than 0.")

        if max_machines < min_machines:
            raise ValueError("`max_machines` should be greater "
                             "than `min_machines`.")

        super().__init__(machine_type=machine_type,
                         data_disk_gb=data_disk_gb,
                         register=register)
        self.min_machines = min_machines
        self.max_machines = max_machines
        self._active_machines = min_machines
        self.__is_elastic = True
        self.spot = spot

        if self.register:
            logging.info("Registering ElasticMachineGroup configurations:")
            self._register_machine_group(min_vms=self.min_machines,
                                         max_vms=self.max_machines,
                                         is_elastic=self.__is_elastic,
                                         num_vms=self._active_machines,
                                         spot=self.spot)

    @classmethod
    def from_api_response(cls, resp: dict):
        machine_group = super().from_api_response(resp)
        machine_group.spot = bool(resp["spot"])
        machine_group.max_machines = int(resp["max_vms"])
        machine_group.min_machines = int(resp["min_vms"])
        machine_group.__dict__["_active_machines"] = int(resp["num_vms"])
        return machine_group

    def active_machines_to_str(self) -> str:
        """Returns a string representation of the 
        number of machines currently running.
        """
        return f"{self._active_machines}/{self.max_machines} (max)"

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}(name=\"{self.name}\")"

    def __str__(self):
        return f"Elastic Machine Group {self.name} with {self.machine_type} " \
             "machines"

    def start(self):
        """Starts minimum number of machines."""
        return super().start(num_vms=self.min_machines,
                             min_vms=self.min_machines,
                             max_vms=self.max_machines,
                             is_elastic=self.__is_elastic,
                             spot=self.spot)

    def terminate(self):
        """Terminates all machines of the machine group."""
        return super().terminate(num_vms=self.min_machines,
                                 min_vms=self.min_machines,
                                 max_vms=self.max_machines,
                                 is_elastic=self.__is_elastic,
                                 spot=self.spot)

    def _log_machine_group_info(self):
        super()._log_machine_group_info()
        logging.info("> Maximum number of machines: %s", self.max_machines)
        logging.info("> Minimum number of machines: %s", self.min_machines)
        logging.info("> Spot: %s", self.spot)
        self.estimate_cloud_cost()

    def estimate_cloud_cost(self):
        """Estimates a cost per hour of min and max machines in US dollars.

        these are the estimted costs of having minimum and the
        maximum number of machines up in the cloud. The final cost will vary
        depending on the total usage of the machines."""
        cost_per_machine = super()._get_estimated_cost(self.spot)
        logging.info(
            "Note: these are the estimated costs of having minimum and the "
            "maximum number of machines up in the cloud. The final cost will "
            "vary depending on the total usage of the machines.")
        logging.info(
            "> Minimum estimated cloud cost of elastic machine group: "
            "%.3f $/h.", cost_per_machine * self.min_machines)
        logging.info(
            "> Maximum estimated cloud cost of elastic machine group:"
            " %.3f $/h.", cost_per_machine * self.max_machines)



"""Client for Inductiva's web API."""
import os
import sys
import logging
import contextvars

import absl

from . import api
from . import simulators
from . import resources
from . import storage
from . import utils
from . import tasks
from . import logs

logs.setup()

api_url = os.environ.get("INDUCTIVA_API_URL", "https://api.inductiva.ai")
_output_dir = contextvars.ContextVar("INDUCTIVA_OUTPUT_DIR")
_output_dir.set(os.environ.get("INDUCTIVA_OUTPUT_DIR", "inductiva_output"))
api_key = os.environ.get("INDUCTIVA_API_KEY")
_checked_key = False

absl.logging.set_verbosity(absl.logging.INFO)

# Disable urllib3 warnings.
# TODO: Verify and fix the appearance of this warning.
urllib3_logger = logging.getLogger("urllib3.connectionpool")
urllib3_logger.setLevel(logging.CRITICAL)

__version__ = "0.4.5"


def set_output_dir(new_output_dir):
    """Sets the value of `inductiva._output_dir` to `new_output_dir`"""
    _output_dir.set(new_output_dir)


def get_output_dir():
    """Returns the value of inductiva._output_dir"""
    return _output_dir.get()


def _check_for_available_package_update():
    # pylint: disable=import-outside-toplevel
    from .localization import translator as __
    import urllib3
    import json

    new_version = __version__

    try:
        http = urllib3.PoolManager()
        resp = http.request("GET",
                            "https://pypi.org/pypi/inductiva/json",
                            headers={"Accept": "application/json"})
        json_response = json.loads(resp.data)
        new_version = json_response["info"]["version"]

    except Exception as ex:  # pylint: disable=broad-exception-caught
        logging.warning(__("failed-update-check", ex), exc_info=True)

    if new_version != __version__:
        msg = __("upgrade-available", new_version, __version__)
        print(msg, file=sys.stderr)


def _check_key():
    global _checked_key

    if not _checked_key and utils.format_utils.getenv_bool(
            "GITHUB_ACTIONS", False) is not True:
        api.methods.validate_api_key(api_key)
        _checked_key = True


_check_for_available_package_update()


def _supports_ansi():
    """Checks if we support ansi formatting for colors and bolds"""
    user_disable_ansi = utils.format_utils.getenv_bool("INDUCTIVA_DISABLE_ANSI",
                                                       False)
    if sys.platform.startswith("win"):
        return "TERM" in os.environ and os.environ[
            "TERM"] == "xterm" and not user_disable_ansi
    return hasattr(sys.stdout,
                   "isatty") and sys.stdout.isatty() and not user_disable_ansi


_ansi_enabled = _supports_ansi()

_check_key()



"""Dummy simulator for demonstrating and testing purposes.

This simulator acts as an echo of the arguments passed and can be used to
test the commands passed to the simulator.
"""

from typing import Optional

from inductiva import types, tasks, simulators


class DummySimulator(simulators.Simulator):
    """Dummy simulator for demo and testing purposes."""

    def __init__(self):
        super().__init__()
        self.api_method_name = "tester.echo.run_simulation"

    def run(self,
            input_dir: types.Path,
            input_filename: str,
            sleep_time: Optional[float] = 1,
            on: Optional[types.ComputationalResources] = None,
            storage_dir: Optional[types.Path] = "",
            extra_metadata: Optional[dict] = None) -> tasks.Task:
        """Run a dummy simulation that echo's to a file.
        
        Args: 
            input_dir: Path to directory with simulation input files.
            input_filename: Name of the test input file.
            sleep_time: Time to sleep before running the commands.
            extra_metadata: Extra metadata to be sent to the backend.
        """

        return super().run(input_dir,
                           on=on,
                           input_filename=input_filename,
                           sleep_time=sleep_time,
                           storage_dir=storage_dir,
                           extra_metadata=extra_metadata)



# coding: utf-8
"""


    Generated by: https://openapi-generator.tech
"""

from dataclasses import dataclass
import typing_extensions
import urllib3
from urllib3._collections import HTTPHeaderDict

from inductiva.client import api_client, exceptions
from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401

from inductiva.client.model.machine_type_response import MachineTypeResponse
from inductiva.client.model.http_validation_error import HTTPValidationError

from . import path

# Query params
NumCpusSchema = schemas.IntSchema


class SpotSchema(
        schemas.ComposedSchema,):

    class MetaOapg:

        class any_of_0(schemas.EnumBase, schemas.StrSchema):

            class MetaOapg:
                enum_value_to_name = {
                    "t": "T",
                    "f": "F",
                }

            @schemas.classproperty
            def T(cls):
                return cls("t")

            @schemas.classproperty
            def F(cls):
                return cls("f")

        class any_of_1(schemas.EnumBase, schemas.BoolSchema):

            class MetaOapg:
                enum_value_to_name = {
                    schemas.BoolClass.TRUE: "TRUE",
                    schemas.BoolClass.FALSE: "FALSE",
                }

            @schemas.classproperty
            def TRUE(cls):
                return cls(True)

            @schemas.classproperty
            def FALSE(cls):
                return cls(False)

        @classmethod
        @functools.lru_cache()
        def any_of(cls):
            # we need this here to make our import statements work
            # we must store _composed_schemas in here so the code is only run
            # when we invoke this method. If we kept this at the class
            # level we would get an error because the class level
            # code would be run when this module is imported, and these composed
            # classes don't exist yet because their module has not finished
            # loading
            return [
                cls.any_of_0,
                cls.any_of_1,
            ]

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
            str,
            date,
            datetime,
            uuid.UUID,
            int,
            float,
            decimal.Decimal,
            bool,
            None,
            list,
            tuple,
            bytes,
            io.FileIO,
            io.BufferedReader,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'SpotSchema':
        return super().__new__(
            cls,
            *_args,
            _configuration=_configuration,
            **kwargs,
        )


RamGbSchema = schemas.IntSchema


class ProviderSchema(schemas.EnumBase, schemas.StrSchema):

    class MetaOapg:
        enum_value_to_name = {
            "GCP": "GCP",
        }

    @schemas.classproperty
    def GCP(cls):
        return cls("GCP")


RequestRequiredQueryParams = typing_extensions.TypedDict(
    'RequestRequiredQueryParams', {
        'num_cpus':
            typing.Union[
                NumCpusSchema,
                decimal.Decimal,
                int,
            ],
        'spot':
            typing.Union[
                SpotSchema,
                dict,
                frozendict.frozendict,
                str,
                date,
                datetime,
                uuid.UUID,
                int,
                float,
                decimal.Decimal,
                bool,
                None,
                list,
                tuple,
                bytes,
                io.FileIO,
                io.BufferedReader,
            ],
    })
RequestOptionalQueryParams = typing_extensions.TypedDict(
    'RequestOptionalQueryParams', {
        'ram_gb': typing.Union[
            RamGbSchema,
            decimal.Decimal,
            int,
        ],
        'provider': typing.Union[
            ProviderSchema,
            str,
        ],
    },
    total=False)


class RequestQueryParams(RequestRequiredQueryParams,
                         RequestOptionalQueryParams):
    pass


request_query_num_cpus = api_client.QueryParameter(
    name="num_cpus",
    style=api_client.ParameterStyle.FORM,
    schema=NumCpusSchema,
    required=True,
    explode=True,
)
request_query_spot = api_client.QueryParameter(
    name="spot",
    style=api_client.ParameterStyle.FORM,
    schema=SpotSchema,
    required=True,
    explode=True,
)
request_query_ram_gb = api_client.QueryParameter(
    name="ram_gb",
    style=api_client.ParameterStyle.FORM,
    schema=RamGbSchema,
    explode=True,
)
request_query_provider = api_client.QueryParameter(
    name="provider",
    style=api_client.ParameterStyle.FORM,
    schema=ProviderSchema,
    explode=True,
)
_auth = [
    'APIKeyHeader',
]
SchemaFor200ResponseBodyApplicationJson = MachineTypeResponse


@dataclass
class ApiResponseFor200(api_client.ApiResponse):
    response: urllib3.HTTPResponse
    body: typing.Union[
        SchemaFor200ResponseBodyApplicationJson,
    ]
    headers: schemas.Unset = schemas.unset


_response_for_200 = api_client.OpenApiResponse(
    response_cls=ApiResponseFor200,
    content={
        'application/json':
            api_client.MediaType(schema=SchemaFor200ResponseBodyApplicationJson
                                ),
    },
)
SchemaFor422ResponseBodyApplicationJson = HTTPValidationError


@dataclass
class ApiResponseFor422(api_client.ApiResponse):
    response: urllib3.HTTPResponse
    body: typing.Union[
        SchemaFor422ResponseBodyApplicationJson,
    ]
    headers: schemas.Unset = schemas.unset


_response_for_422 = api_client.OpenApiResponse(
    response_cls=ApiResponseFor422,
    content={
        'application/json':
            api_client.MediaType(schema=SchemaFor422ResponseBodyApplicationJson
                                ),
    },
)
_status_code_to_response = {
    '200': _response_for_200,
    '422': _response_for_422,
}
_all_accept_content_types = ('application/json',)


class BaseApi(api_client.Api):

    @typing.overload
    def _get_machine_type_oapg(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
            ApiResponseFor200,
    ]:
        ...

    @typing.overload
    def _get_machine_type_oapg(
        self,
        skip_deserialization: typing_extensions.Literal[True],
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization:
        ...

    @typing.overload
    def _get_machine_type_oapg(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
            ApiResponseFor200,
            api_client.ApiResponseWithoutDeserialization,
    ]:
        ...

    def _get_machine_type_oapg(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        """
        Get Machine Type
        :param skip_deserialization: If true then api_response.response will be set but
            api_response.body and api_response.headers will not be deserialized into schema
            class instances
        """
        self._verify_typed_dict_inputs_oapg(RequestQueryParams, query_params)
        used_path = path.value

        prefix_separator_iterator = None
        for parameter in (
                request_query_num_cpus,
                request_query_spot,
                request_query_ram_gb,
                request_query_provider,
        ):
            parameter_data = query_params.get(parameter.name, schemas.unset)
            if parameter_data is schemas.unset:
                continue
            if prefix_separator_iterator is None:
                prefix_separator_iterator = parameter.get_prefix_separator_iterator(
                )
            serialized_data = parameter.serialize(parameter_data,
                                                  prefix_separator_iterator)
            for serialized_value in serialized_data.values():
                used_path += serialized_value

        _headers = HTTPHeaderDict()
        # TODO add cookie handling
        if accept_content_types:
            for accept_content_type in accept_content_types:
                _headers.add('Accept', accept_content_type)

        response = self.api_client.call_api(
            resource_path=used_path,
            method='get'.upper(),
            headers=_headers,
            auth_settings=_auth,
            stream=stream,
            timeout=timeout,
        )

        if skip_deserialization:
            api_response = api_client.ApiResponseWithoutDeserialization(
                response=response)
        else:
            response_for_status = _status_code_to_response.get(
                str(response.status))
            if response_for_status:
                api_response = response_for_status.deserialize(
                    response, self.api_client.configuration)
            else:
                api_response = api_client.ApiResponseWithoutDeserialization(
                    response=response)

        if not 200 <= response.status <= 299:
            raise exceptions.ApiException(status=response.status,
                                          reason=response.reason,
                                          api_response=api_response)

        return api_response


class GetMachineType(BaseApi):
    # this class is used by api classes that refer to endpoints with operationId fn names

    @typing.overload
    def get_machine_type(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
            ApiResponseFor200,
    ]:
        ...

    @typing.overload
    def get_machine_type(
        self,
        skip_deserialization: typing_extensions.Literal[True],
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization:
        ...

    @typing.overload
    def get_machine_type(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
            ApiResponseFor200,
            api_client.ApiResponseWithoutDeserialization,
    ]:
        ...

    def get_machine_type(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        return self._get_machine_type_oapg(
            query_params=query_params,
            accept_content_types=accept_content_types,
            stream=stream,
            timeout=timeout,
            skip_deserialization=skip_deserialization)


class ApiForget(BaseApi):
    # this class is used by api classes that refer to endpoints by path and http method names

    @typing.overload
    def get(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
            ApiResponseFor200,
    ]:
        ...

    @typing.overload
    def get(
        self,
        skip_deserialization: typing_extensions.Literal[True],
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization:
        ...

    @typing.overload
    def get(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
            ApiResponseFor200,
            api_client.ApiResponseWithoutDeserialization,
    ]:
        ...

    def get(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        return self._get_machine_type_oapg(
            query_params=query_params,
            accept_content_types=accept_content_types,
            stream=stream,
            timeout=timeout,
            skip_deserialization=skip_deserialization)



"""Available machine types and their number of cores."""

import os
import yaml

MACHINE_TYPES_FILE = os.path.join(os.path.dirname(__file__),
                                  "machine_types.yaml")


def get_available():
    """Gets the available machine types.
    
    Currently, this fetches the available machine types from a yaml file
    into a dictionary that can now be used to parse the information where
    needed."""

    with open(MACHINE_TYPES_FILE, "r", encoding="utf-8") as file:
        machine_types = yaml.safe_load(file)

    return machine_types


def list_available_machines():
    """List all available machines types."""

    resources_available = get_available()

    machine_types = []

    # Fetch the provider information
    for _, provider_resources in resources_available.items():
        # Fetch the available machine CPU series
        for cpu_series, series_info in provider_resources["cpu-series"].items():
            # Fetch the available RAM types and vCPUs info
            for ram_type, type_info in series_info["types"].items():
                vcpus = type_info["vcpus"]
                machine_types.extend(
                    [f"{cpu_series}-{ram_type}-{vcpu}" for vcpu in vcpus])
                if type_info["lssd"]:
                    for vcpu in vcpus:
                        machine_types.append(
                            f"{cpu_series}-{ram_type}-{vcpu}-lssd")

    return tuple(machine_types)



import typing_extensions

from inductiva.client.paths import PathValues
from inductiva.client.apis.paths.tasks_auth import TasksAuth
from inductiva.client.apis.paths.tasks_submit import TasksSubmit
from inductiva.client.apis.paths.tasks_task_id_input import TasksTaskIdInput
from inductiva.client.apis.paths.tasks_task_id import TasksTaskId
from inductiva.client.apis.paths.tasks import Tasks
from inductiva.client.apis.paths.tasks_task_id_status import TasksTaskIdStatus
from inductiva.client.apis.paths.tasks_task_id_output_list import TasksTaskIdOutputList
from inductiva.client.apis.paths.tasks_task_id_output import TasksTaskIdOutput
from inductiva.client.apis.paths.tasks_task_id_kill import TasksTaskIdKill
from inductiva.client.apis.paths.admin_users import AdminUsers
from inductiva.client.apis.paths.admin_users_email_api_key import AdminUsersEmailApiKey
from inductiva.client.apis.paths.admin_users_username_tasks import AdminUsersUsernameTasks
from inductiva.client.apis.paths.admin_groups import AdminGroups
from inductiva.client.apis.paths.admin_active_tasks import AdminActiveTasks
from inductiva.client.apis.paths.executer_tracker_register import ExecuterTrackerRegister
from inductiva.client.apis.paths.compute_group import ComputeGroup
from inductiva.client.apis.paths.compute_type import ComputeType
from inductiva.client.apis.paths.compute_group_start import ComputeGroupStart
from inductiva.client.apis.paths.compute_price import ComputePrice
from inductiva.client.apis.paths.compute_groups import ComputeGroups
from inductiva.client.apis.paths.compute_group_status import ComputeGroupStatus
from inductiva.client.apis.paths.storage_size import StorageSize
from inductiva.client.apis.paths.storage_contents import StorageContents
from inductiva.client.apis.paths.version import Version
from inductiva.client.apis.paths.version_check import VersionCheck

PathToApi = typing_extensions.TypedDict(
    'PathToApi', {
        PathValues.TASKS_AUTH: TasksAuth,
        PathValues.TASKS_SUBMIT: TasksSubmit,
        PathValues.TASKS_TASK_ID_INPUT: TasksTaskIdInput,
        PathValues.TASKS_TASK_ID: TasksTaskId,
        PathValues.TASKS: Tasks,
        PathValues.TASKS_TASK_ID_STATUS: TasksTaskIdStatus,
        PathValues.TASKS_TASK_ID_OUTPUT_LIST: TasksTaskIdOutputList,
        PathValues.TASKS_TASK_ID_OUTPUT: TasksTaskIdOutput,
        PathValues.TASKS_TASK_ID_KILL: TasksTaskIdKill,
        PathValues.ADMIN_USERS: AdminUsers,
        PathValues.ADMIN_USERS_EMAIL_API_KEY: AdminUsersEmailApiKey,
        PathValues.ADMIN_USERS_USERNAME_TASKS: AdminUsersUsernameTasks,
        PathValues.ADMIN_GROUPS: AdminGroups,
        PathValues.ADMIN_ACTIVE_TASKS: AdminActiveTasks,
        PathValues.EXECUTERTRACKER_REGISTER: ExecuterTrackerRegister,
        PathValues.COMPUTE_GROUP: ComputeGroup,
        PathValues.COMPUTE_TYPE: ComputeType,
        PathValues.COMPUTE_GROUP_START: ComputeGroupStart,
        PathValues.COMPUTE_PRICE: ComputePrice,
        PathValues.COMPUTE_GROUPS: ComputeGroups,
        PathValues.COMPUTE_GROUP_STATUS: ComputeGroupStatus,
        PathValues.STORAGE_SIZE: StorageSize,
        PathValues.STORAGE_CONTENTS: StorageContents,
        PathValues.VERSION: Version,
        PathValues.VERSIONCHECK: VersionCheck,
    })

path_to_api = PathToApi({
    PathValues.TASKS_AUTH: TasksAuth,
    PathValues.TASKS_SUBMIT: TasksSubmit,
    PathValues.TASKS_TASK_ID_INPUT: TasksTaskIdInput,
    PathValues.TASKS_TASK_ID: TasksTaskId,
    PathValues.TASKS: Tasks,
    PathValues.TASKS_TASK_ID_STATUS: TasksTaskIdStatus,
    PathValues.TASKS_TASK_ID_OUTPUT_LIST: TasksTaskIdOutputList,
    PathValues.TASKS_TASK_ID_OUTPUT: TasksTaskIdOutput,
    PathValues.TASKS_TASK_ID_KILL: TasksTaskIdKill,
    PathValues.ADMIN_USERS: AdminUsers,
    PathValues.ADMIN_USERS_EMAIL_API_KEY: AdminUsersEmailApiKey,
    PathValues.ADMIN_USERS_USERNAME_TASKS: AdminUsersUsernameTasks,
    PathValues.ADMIN_GROUPS: AdminGroups,
    PathValues.ADMIN_ACTIVE_TASKS: AdminActiveTasks,
    PathValues.EXECUTERTRACKER_REGISTER: ExecuterTrackerRegister,
    PathValues.COMPUTE_GROUP: ComputeGroup,
    PathValues.COMPUTE_TYPE: ComputeType,
    PathValues.COMPUTE_GROUP_START: ComputeGroupStart,
    PathValues.COMPUTE_PRICE: ComputePrice,
    PathValues.COMPUTE_GROUPS: ComputeGroups,
    PathValues.COMPUTE_GROUP_STATUS: ComputeGroupStatus,
    PathValues.STORAGE_SIZE: StorageSize,
    PathValues.STORAGE_CONTENTS: StorageContents,
    PathValues.VERSION: Version,
    PathValues.VERSIONCHECK: VersionCheck,
})



"""List the tasks information via CLI."""
import argparse

from inductiva import tasks, utils
from inductiva.client import models


def list_tasks(args):
    """ List the last user's tasks. 

    Lists based on the flags (task_id, last_n).
    It can print the last N tasks (default value is 5)
    or a single task with a specific ID
    """
    if args.task_id is not None:
        task_list = [tasks.Task(args.task_id)]

    else:
        last_n = 5 if args.last_n is None else args.last_n
        task_list = tasks.get(last_n=last_n)

    if not task_list:
        print("No tasks found.")
        return 1

    table_dict = tasks.to_dict(task_list)

    emph_formatter = utils.format_utils.get_ansi_formatter()

    def color_formater(status):
        if status == models.TaskStatusCode.SUCCESS:
            return emph_formatter(status, utils.format_utils.Emphasis.GREEN)
        elif status in tasks.Task.FAILED_STATUSES:
            return emph_formatter(status, utils.format_utils.Emphasis.RED)
        return status

    formatters = {
        "Submitted": [utils.format_utils.datetime_formatter,],
        "Started": [utils.format_utils.datetime_formatter,],
        "Status": [color_formater]
    }

    header_formatters = [
        lambda x: emph_formatter(x.upper(), utils.format_utils.Emphasis.BOLD)
    ]

    print(
        utils.format_utils.get_tabular_str(table_dict,
                                           formatters=formatters,
                                           header_formatters=header_formatters))
    return 0


def register(parser):
    """Register the list user's tasks command."""

    subparser = parser.add_parser("list",
                                  help="List tasks.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = ("The `inductiva tasks list` command provides "
                             "an overview of your tasks on the platform.\n"
                             "It lists the most recent tasks or a specific "
                             "task by ID.\n"
                             "You can control the number of tasks listed with "
                             "the '-n' or '--last-n' option.\n")

    group = subparser.add_mutually_exclusive_group()
    group.add_argument("-n",
                       "--last-n",
                       type=int,
                       help="List last N tasks. Default: 5.")
    group.add_argument("-id",
                       "--task-id",
                       type=str,
                       help="List a task with a specific ID.")

    subparser.set_defaults(func=list_tasks)



"""SplisHSPlasH simulator module of the API."""
from typing import Optional

from inductiva import types, tasks, simulators


class SplishSplash(simulators.Simulator):
    """Class to invoke a generic SPlisHSPlasH simulation on the API."""

    def __init__(self):
        super().__init__()
        self.api_method_name = "sph.splishsplash.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        sim_config_filename: str,
        on: Optional[types.ComputationalResources] = None,
        storage_dir: Optional[types.Path] = "",
        extra_metadata: Optional[dict] = None,
    ) -> tasks.Task:
        """Run the SPlisHSPlasH simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            sim_config_filename: Name of the simulation configuration file.
            on: The computational resource to launch the simulation on. If None
                the simulation is submitted to a machine in the default pool.
            storage_dir: Directory for storing simulation results.
        Returns:
            Task object representing the simulation task.
        """
        return super().run(
            input_dir,
            on=on,
            input_filename=sim_config_filename,
            storage_dir=storage_dir,
            extra_metadata=extra_metadata,
        )



# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from inductiva.client.apis.path_to_api import path_to_api

import enum


class PathValues(str, enum.Enum):
    TASKS_AUTH = "/tasks/auth"
    TASKS_SUBMIT = "/tasks/submit"
    TASKS_TASK_ID_INPUT = "/tasks/{task_id}/input"
    TASKS_TASK_ID = "/tasks/{task_id}"
    TASKS = "/tasks"
    TASKS_TASK_ID_STATUS = "/tasks/{task_id}/status"
    TASKS_TASK_ID_OUTPUT_LIST = "/tasks/{task_id}/output/list"
    TASKS_TASK_ID_OUTPUT = "/tasks/{task_id}/output"
    TASKS_TASK_ID_KILL = "/tasks/{task_id}/kill"
    ADMIN_USERS = "/admin/users"
    ADMIN_USERS_EMAIL_API_KEY = "/admin/users/{email}/api_key"
    ADMIN_USERS_USERNAME_TASKS = "/admin/users/{username}/tasks"
    ADMIN_GROUPS = "/admin/groups"
    ADMIN_ACTIVE_TASKS = "/admin/active_tasks"
    EXECUTERTRACKER_REGISTER = "/executer-tracker/register"
    COMPUTE_GROUP = "/compute/group"
    COMPUTE_TYPE = "/compute/type"
    COMPUTE_GROUP_START = "/compute/group/start"
    COMPUTE_PRICE = "/compute/price"
    COMPUTE_GROUPS = "/compute/groups"
    COMPUTE_GROUP_STATUS = "/compute/group_status"
    STORAGE_SIZE = "/storage/size"
    STORAGE_CONTENTS = "/storage/contents"
    VERSION = "/version"
    VERSIONCHECK = "/version-check"
